﻿using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Encodings.BinaryVectorEncoding;
using HeuristicLab.Encodings.PermutationEncoding;
using HeuristicLab.Optimization;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Problems.TravelingSalesman;

namespace HeuristicLab.Networks.IntegratedOptimization {
  [Item("Tour Profit Problem", "Represents a problem whose objective is to optimize a TTP tour for fixed TTP loot.")]
  [Creatable(CreatableAttribute.Categories.Problems, Priority = 999)]
  [StorableClass]
  public class TourProfitProblem : SingleObjectiveBasicProblem<PermutationEncoding> {
    public override bool Maximization {
      get { return true; }
    }

    [Storable]
    public TravelingSalesmanProblem Tsp { get; set; }
    [Storable]
    public BinaryKnapsackProblem Ksp { get; set; }
    [Storable]
    public BinaryVector FixedKspSolution { get; set; }
    [Storable]
    public int[] Availability { get; set; }
    [Storable]
    public double RentingRatio { get; set; }
    [Storable]
    public double MinSpeed { get; set; }
    [Storable]
    public double MaxSpeed { get; set; }

    [StorableConstructor]
    protected TourProfitProblem(bool deserializing) : base(deserializing) { }
    protected TourProfitProblem(TourProfitProblem original, Cloner cloner) : base(original, cloner) {
      Tsp = cloner.Clone(original.Tsp);
      Ksp = cloner.Clone(original.Ksp);
      FixedKspSolution = cloner.Clone(original.FixedKspSolution);
      Availability = original.Availability != null ? (int[])original.Availability.Clone() : null;
      RentingRatio = original.RentingRatio;
      MinSpeed = original.MinSpeed;
      MaxSpeed = original.MaxSpeed;
    }
    public TourProfitProblem() : base() {
      Encoding.Length = 5;
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new TourProfitProblem(this, cloner);
    }

    public override double Evaluate(Individual individual, IRandom random) {
      return TtpUtils.EvaluateTtp(Tsp, individual.Permutation().CloneAsArray(),
                                  Ksp, FixedKspSolution.CloneAsArray(),
                                  Availability, RentingRatio, MinSpeed, MaxSpeed);
    }

    public override IEnumerable<Individual> GetNeighbors(Individual individual, IRandom random) {
      while (true) {
        var neighbor = individual.Copy();
        switch (random.Next(7)) {
          case 0: InsertionManipulator.Apply(random, neighbor.Permutation()); break;
          case 1: InversionManipulator.Apply(random, neighbor.Permutation()); break;
          case 2: ScrambleManipulator.Apply(random, neighbor.Permutation()); break;
          case 3: Swap2Manipulator.Apply(random, neighbor.Permutation()); break;
          case 4: Swap3Manipulator.Apply(random, neighbor.Permutation()); break;
          case 5: TranslocationInversionManipulator.Apply(random, neighbor.Permutation()); break;
          case 6: TranslocationManipulator.Apply(random, neighbor.Permutation()); break;
        }
        yield return neighbor;
      }
    }

    public override void Analyze(Individual[] individuals, double[] qualities, ResultCollection results, IRandom random) {
      base.Analyze(individuals, qualities, results, random);
      var orderedIndividuals = individuals.Zip(qualities, (i, q) => new { Individual = i, Quality = q }).OrderBy(z => z.Quality);
      var best = Maximization ? orderedIndividuals.Last().Individual : orderedIndividuals.First().Individual;

      if (!results.ContainsKey("Best TSP Solution"))
        results.Add(new Result("Best TSP Solution", typeof(Permutation)));
      results["Best TSP Solution"].Value = (IItem)best.Permutation().Clone();
    }
  }
}
